// Copyright (c) 2006 David Drysdale.  See license.txt for license info
import java.awt.*;

class AngleController implements ClickConsumer, DragConsumer, Control {
  final static int SIZE = 50;
  final static int MARGIN = 5;
  final static int LINEWIDTH = 12;
  final static int HALFARC = 8;
  final static int INNERRADIUS = ((SIZE/2)-MARGIN-LINEWIDTH);
  final static int OUTERRADIUS = ((SIZE/2)-MARGIN);
  final static Color COLOR0=Color.black;
  final static Color COLOR1NOK=Color.red;
  final static Color COLOR1OK=Color.green;
  final static Color COLOR2=Color.white;
  final static double RADIANS=(360.0 / (2*Math.PI));
  int centrex_;
  int centrey_;
  Control nextControl_;
  ClickConsumer nextClick_;
  DragConsumer nextDrag_;
  Triggerable trigger_;
  SharedDouble value_;
  double minVal_;
  double maxVal_;
  boolean inClick_ = false;
  static boolean verbose_ = false;

  public AngleController(SharedDouble val, double min, double max, Triggerable trigger, int x, int y) {
    value_ = val;
    minVal_ = min;
    maxVal_ = max;
    trigger_ = trigger;
    centrex_ = x + (SIZE/2);
    centrey_ = y + (SIZE/2);
  }
  public void display(Drawable2D c, Graphics g) {
    // Control is in a box SIZE x SIZE
    g.setColor(COLOR1NOK);
    g.fillOval(centrex_ - OUTERRADIUS, centrey_ - OUTERRADIUS, 2*OUTERRADIUS, 2*OUTERRADIUS);
    g.setColor(COLOR1OK);
    g.fillArc(centrex_ - OUTERRADIUS, centrey_ - OUTERRADIUS, 2*OUTERRADIUS, 2*OUTERRADIUS,
              90-(int)(maxVal_*RADIANS), (int)((maxVal_-minVal_)*RADIANS));
    g.setColor(COLOR0);
    g.fillOval(centrex_ - INNERRADIUS, centrey_ - INNERRADIUS, 2*INNERRADIUS, 2*INNERRADIUS);
    g.setColor(COLOR2);
    g.fillArc(centrex_ - OUTERRADIUS, centrey_ - OUTERRADIUS, 2*OUTERRADIUS, 2*OUTERRADIUS,
              90-(int)(value_.doubleValue()*RADIANS)-HALFARC, 2*HALFARC);
    if (nextControl_ != null) {
      nextControl_.display(c,g);
    }
  }
  private boolean processPt(int rawx, int rawy, boolean checkDist) {
    boolean processed = false;
    double dist = Math.sqrt((rawx-centrex_)*(rawx-centrex_) + (rawy-centrey_)*(rawy-centrey_));

    if ((!checkDist) || (dist <= (SIZE/2))) {
      if (verbose_) System.out.println("click at ("+rawx+","+rawy+") dist="+dist+" in "+this);
      double deltax = rawx - centrex_;
      double deltay = rawy - centrey_;
      double angle = Math.atan2(deltax, -deltay);
      if (angle < minVal_) angle=minVal_;
      if (angle > maxVal_) angle=maxVal_;
      value_.setDoubleValue(angle);
      processed = true;
    }
    if (processed) {
      // Trigger a regeneration of the frames and a redisplay
      trigger_.generateFrames();
      trigger_.trigger();
    }
    return processed;
  }
  public boolean click(double x, double y, int rawx, int rawy) {
    boolean processed = processPt(rawx, rawy, true);
    inClick_ = processed;
    if (!processed && (nextClick_ != null)) {
      // Pass it on down the chain
      processed = nextClick_.click(x,y,rawx,rawy);     
    }
    return processed;
  }
  public boolean release(double x, double y, int rawx, int rawy) {
    inClick_ = false;
    if (nextClick_ != null) {
      return nextClick_.release(x,y,rawx,rawy);     
    } else {
      return false;
    }
  }
  public boolean drag(double x, double y, int rawx, int rawy) {
    if (inClick_) processPt(rawx, rawy, false);
    boolean processed = inClick_;
    if (!processed && (nextDrag_ != null)) {
      // Pass it on down the chain
      processed = nextDrag_.drag(x,y,rawx,rawy);
    }
    return processed;
  }
  public boolean ctrlDrag(double x, double y, int rawx, int rawy) {
    if (nextDrag_ != null) {
      return nextDrag_.ctrlDrag(x,y,rawx,rawy);
    } else {
      return false;
    }
  }
};