// Copyright (c) 2006 David Drysdale.  See license.txt for license info
import java.applet.Applet;
import java.awt.*;

class AnimateThread extends Thread {
  // Owning applet, which we call back into with time updates
  private Triggerable trigger_;
  // Whether looping or not
  private Checkbox repeatCheckbox_;
  // The total amount of world-time for the animation when looping
  private double durationt_;
  // The amount of world-time when not looping
  private double nonloopdurationt_;
  // Current world-time (reference to an object elsewhere)
  private SharedDouble curt_;
  // Whether thread is finished processing
  private boolean done_ = false;
  // Rate at which world-time passes in real time
  private double speed_ = 4;

  public AnimateThread(Triggerable trigger, Checkbox repeatCheckbox, double duration, double nonloop, SharedDouble curt) {
    trigger_ = trigger;
    repeatCheckbox_ = repeatCheckbox;
    durationt_ = duration;
    nonloopdurationt_ = nonloop;
    curt_ = curt;
  }
  public boolean done() { return done_; }
  public void setSpeed(double speed) {speed_=speed;}
  public double getSpeed() {return speed_;}
  public synchronized void run() {
    boolean firstIn = true;
    done_ = false;
    do {
      if (!firstIn) {
        // Gone back to the beginning
        curt_.setDoubleValue(0.0);
      }
      
      // Figure out the time for the whole sequence, assuming a loop. We use this
      // for calculations even when not looping, so that toggling the loop flag
      // doesn't change any numbers
      long durationms = (long)(1000.0*(durationt_/speed_));

      // Figure out the start time for this animation
      long startms = System.currentTimeMillis();
      startms -= (long)((curt_.doubleValue() * (double)durationms) / durationt_);

      while ((curt_.doubleValue() <= (repeatCheckbox_.getState()?durationt_:nonloopdurationt_)) && !done_) {
        // Frob the trigger because things have changed
        trigger_.trigger();
        try {
          Thread.sleep(50); 
        } catch (InterruptedException e) {}
        // Move current time forward by the appropriate amount
        long deltams = System.currentTimeMillis() - startms;
        curt_.setDoubleValue(durationt_ * ((double)deltams / (double)durationms));
      }
      firstIn = false;
    } while (repeatCheckbox_.getState() && !done_);
    done_ = true;
    // Notify the trigger that we're done
    trigger_.animThreadDone();
  }
  public void term() {
    done_ = true;
  }
};