// Copyright (c) 2006 David Drysdale.  See license.txt for license info
import java.awt.*;

class Animator {

  private AnimCanvas plot_;
  private Dimension plotDim_;
  private Label timeLabel_;

  private Line3DSequence frames_;
  private double duration_;

  public Animator(AnimCanvas plot, Label timeLabel, Line3DSequence frames) {
    plot_ = plot;
    plotDim_ = new Dimension();
    plotDim_ = plot.size();
    frames_ = frames;
    duration_ = frames_.totalDuration();
    timeLabel_ = timeLabel;
  }

  public double totalDuration() {return duration_;}
  public double activeDuration(boolean repeat) {
    double duration = totalDuration();
    if (!repeat) duration -= frames_.deltat_[0];
    return duration;
  }
  private Line2D[] interpolate(Line2D[] linesa, Line2D[] linesb, double ratio) {
    if (linesa.length != linesb.length) {
      if (ratio < 0.5) {
        return linesa;
      } else {
        return linesb;
      }
    }
    Line2D[] result = new Line2D[linesa.length];
    for (int ii=0; ii<result.length; ii++) {
      result[ii] = new Line2D();
      result[ii].a.x = linesa[ii].a.x + (ratio * (double)(linesb[ii].a.x - linesa[ii].a.x));
      result[ii].a.y = linesa[ii].a.y + (ratio * (double)(linesb[ii].a.y - linesa[ii].a.y));
      result[ii].b.x = linesa[ii].b.x + (ratio * (double)(linesb[ii].b.x - linesa[ii].b.x));
      result[ii].b.y = linesa[ii].b.y + (ratio * (double)(linesb[ii].b.y - linesa[ii].b.y));
      result[ii].c = linesa[ii].c;
    }
    return result;
  }

  public void drawFrame(double curt, boolean repeat, Camera camera) {
    if (timeLabel_ != null) {
      curt = (double)((int)(10.0*curt))/10.0;
      String tString = new String("t = "+curt);
      timeLabel_.setText(tString);
    }
    
    // Figure out which two frames we are between
    int ii = 0;
    int nextii = 1;
    double totTime = 0.0;
    while (curt > (totTime + frames_.deltat_[nextii])) {
      ii = nextii;
      nextii++;
      if (nextii == frames_.length()) nextii = 0;
      totTime += frames_.deltat_[ii];
    }
    
    // Current frame is between [ii, nextii); the corresponding times
    // are [totTime, totTime + frames_.deltat_[nextii])
    if ((nextii == 0) && !repeat) nextii = ii;
    Line3D[] framebefore = frames_.lines_[ii];
    Line3D[] frameafter = frames_.lines_[nextii];
    Line2D lines[] = interpolate(camera.transform(framebefore),
                                 camera.transform(frameafter),
                                 (curt-totTime)/(frames_.deltat_[nextii]));
    plot_.lines(lines);
    plot_.drawDone();
  }
  public void drawFrame(int frame, double curt, boolean repeat, Camera camera) {
    if (timeLabel_ != null) {
      curt = (double)((int)(10.0*curt))/10.0;
      String tString = new String("t = "+curt);
      timeLabel_.setText(tString);
    }
    Line3D[] frame3d = frames_.lines_[frame];
    Line2D lines[] = camera.transform(frame3d);
    plot_.lines(lines);
    plot_.drawDone();
  }
};