// Copyright (c) 2006 David Drysdale.  See license.txt for license info
class Figure {
  final static double DEGREES=((2*Math.PI) / 360.0);
  
  // Lengths (all in metres)
  double foot_ = 0.22;
  double tibia_ = 0.45;
  double femur_ = 0.47;
  double halfpelvis_ = 0.15;
  double lowerspine_ = 0.29;
  double upperspine_ = 0.29;
  double halfclavicle_ = 0.24;
  double humerus_ = 0.28;
  double ulna_ = 0.26;
  double hand_ = 0.18;
  double neck_ = 0.13;
  double headheight_ = 0.24;
  double headwidth_ = 0.18;

  // Number of lines in stick figure
  public static final int lineCount_ = 22;
  // Number of lines in bounding box
  public static final int bbLineCount_ = 12;
  
  // Small tweak for selectability
  private static final double EPSILON=0.01;

  // Angles, in radians.  Default values of 0.0 are normal positions  
  SharedDouble[] hipforward_;
  SharedDouble[] hipside_;
  SharedDouble[] hiprotate_;
  SharedDouble[] knee_;
  SharedDouble[] ankle_;
  SharedDouble spineforward_;
  SharedDouble spineside_;
  SharedDouble spinerotate_;
  SharedDouble neckforward_;
  SharedDouble neckside_;
  SharedDouble neckrotate_;
  SharedDouble[] shoulderforward_;
  SharedDouble[] shoulderup_;
  SharedDouble[] humerusforward_;
  SharedDouble[] humerusside_;
  SharedDouble[] humerusrotate_;
  SharedDouble[] elbow_;
  SharedDouble[] wristup_;
  SharedDouble[] wristside_;
  
  private void arrayAllocate() {
    hipforward_ = new SharedDouble[2];
    hipside_ = new SharedDouble[2];
    hiprotate_ = new SharedDouble[2];
    knee_ = new SharedDouble[2];
    ankle_ = new SharedDouble[2];
    shoulderforward_ = new SharedDouble[2];
    shoulderup_ = new SharedDouble[2];
    humerusforward_ = new SharedDouble[2];
    humerusside_ = new SharedDouble[2];
    humerusrotate_ = new SharedDouble[2];
    elbow_ = new SharedDouble[2];
    wristup_ = new SharedDouble[2];
    wristside_ = new SharedDouble[2];
  }
  Figure() {
    arrayAllocate();
    hipforward_[0] = new SharedDouble();
    hipside_[0] = new SharedDouble();
    hiprotate_[0] = new SharedDouble();
    knee_[0] = new SharedDouble();
    ankle_[0] = new SharedDouble();
    hipforward_[1] = new SharedDouble();
    hipside_[1] = new SharedDouble();
    hiprotate_[1] = new SharedDouble();
    knee_[1] = new SharedDouble();
    ankle_[1] = new SharedDouble();
    spineforward_ = new SharedDouble();
    spineside_ = new SharedDouble();
    spinerotate_ = new SharedDouble();
    neckforward_ = new SharedDouble();
    neckside_ = new SharedDouble();
    neckrotate_ = new SharedDouble();
    shoulderforward_[0] = new SharedDouble();
    shoulderup_[0] = new SharedDouble();
    humerusforward_[0] = new SharedDouble();
    humerusside_[0] = new SharedDouble();
    humerusrotate_[0] = new SharedDouble();
    elbow_[0] = new SharedDouble();
    wristup_[0] = new SharedDouble();
    wristside_[0] = new SharedDouble();
    shoulderforward_[1] = new SharedDouble();
    shoulderup_[1] = new SharedDouble();
    humerusforward_[1] = new SharedDouble();
    humerusside_[1] = new SharedDouble();
    humerusrotate_[1] = new SharedDouble();
    elbow_[1] = new SharedDouble();
    wristup_[1] = new SharedDouble();
    wristside_[1] = new SharedDouble();
  }
  Figure(Figure copy) {
    arrayAllocate();
    hipforward_[0] = new SharedDouble(copy.hipforward_[0].doubleValue());
    hipside_[0] = new SharedDouble(copy.hipside_[0].doubleValue());
    hiprotate_[0] = new SharedDouble(copy.hiprotate_[0].doubleValue());
    knee_[0] = new SharedDouble(copy.knee_[0].doubleValue());
    ankle_[0] = new SharedDouble(copy.ankle_[0].doubleValue());
    hipforward_[1] = new SharedDouble(copy.hipforward_[1].doubleValue());
    hipside_[1] = new SharedDouble(copy.hipside_[1].doubleValue());
    hiprotate_[1] = new SharedDouble(copy.hiprotate_[1].doubleValue());
    knee_[1] = new SharedDouble(copy.knee_[1].doubleValue());
    ankle_[1] = new SharedDouble(copy.ankle_[1].doubleValue());
    spineforward_ = new SharedDouble(copy.spineforward_.doubleValue());
    spineside_ = new SharedDouble(copy.spineside_.doubleValue());
    spinerotate_ = new SharedDouble(copy.spinerotate_.doubleValue());
    neckforward_ = new SharedDouble(copy.neckforward_.doubleValue());
    neckside_ = new SharedDouble(copy.neckside_.doubleValue());
    neckrotate_ = new SharedDouble(copy.neckrotate_.doubleValue());
    shoulderforward_[0] = new SharedDouble(copy.shoulderforward_[0].doubleValue());
    shoulderup_[0] = new SharedDouble(copy.shoulderup_[0].doubleValue());
    humerusforward_[0] = new SharedDouble(copy.humerusforward_[0].doubleValue());
    humerusside_[0] = new SharedDouble(copy.humerusside_[0].doubleValue());
    humerusrotate_[0] = new SharedDouble(copy.humerusrotate_[0].doubleValue());
    elbow_[0] = new SharedDouble(copy.elbow_[0].doubleValue());
    wristup_[0] = new SharedDouble(copy.wristup_[0].doubleValue());
    wristside_[0] = new SharedDouble(copy.wristside_[0].doubleValue());
    shoulderforward_[1] = new SharedDouble(copy.shoulderforward_[1].doubleValue());
    shoulderup_[1] = new SharedDouble(copy.shoulderup_[1].doubleValue());
    humerusforward_[1] = new SharedDouble(copy.humerusforward_[1].doubleValue());
    humerusside_[1] = new SharedDouble(copy.humerusside_[1].doubleValue());
    humerusrotate_[1] = new SharedDouble(copy.humerusrotate_[1].doubleValue());
    elbow_[1] = new SharedDouble(copy.elbow_[1].doubleValue());
    wristup_[1] = new SharedDouble(copy.wristup_[1].doubleValue());
    wristside_[1] = new SharedDouble(copy.wristside_[1].doubleValue());
  }

  // Joint identifiers
  static final int HIPLEFT = 0;
  static final int HIPRIGHT = 1;
  static final int KNEELEFT = 2;
  static final int KNEERIGHT = 3;
  static final int ANKLELEFT = 4;
  static final int ANKLERIGHT = 5;
  static final int SHOULDERLEFT = 6;
  static final int SHOULDERRIGHT = 7;
  static final int HUMERUSLEFT = 8;
  static final int HUMERUSRIGHT = 9;
  static final int ELBOWLEFT = 10;
  static final int ELBOWRIGHT = 11;
  static final int WRISTLEFT = 12;
  static final int WRISTRIGHT = 13;
  static final int SPINE = 14;
  static final int NECK = 15;
  // Map from joint number to number of degrees of freedom
  static final int[] jointToDOF_ = {3,3, // HIP
                                    1,1, // KNEE
                                    1,1, // ANKLE
                                    2,2, // SHOULDER
                                    3,3, // HUMERUS
                                    1,1, // ELBOW
                                    2,2, // WRIST
                                    3,3}; // SPINE, NECK
  // Map from joint number to line number (always the A-end)
  static final int[] jointToLine_ = {1,5, // HIP
                                     3,6, // KNEE
                                     4,7, // ANKLE
                                     14,18, // SHOULDER
                                     15,19, // HUMERUS
                                     16,20, // ELBOW
                                     17,21, // WRIST
                                     8, 9}; // SPINE, NECK
  // Map from A (B) end of given line to joint number
  static final int[] lineAToJoint_ = {SPINE, HIPLEFT,  // 0 1
                                      HIPLEFT, KNEELEFT, ANKLELEFT, // 2 3 4
                                      HIPRIGHT, KNEERIGHT, ANKLERIGHT, // 5 6 7
                                      SPINE, NECK, NECK, NECK, NECK, NECK, // 8 9 10 11 12 13
                                      SHOULDERLEFT, HUMERUSLEFT, ELBOWLEFT, WRISTLEFT, // 14 15 16 17
                                      SHOULDERRIGHT, HUMERUSRIGHT, ELBOWRIGHT, WRISTRIGHT}; // 18 19 20 21
  static final int[] lineBToJoint_ = {SPINE, HIPRIGHT,  // 0 1
                                      KNEELEFT, ANKLELEFT, ANKLELEFT, // 2 3 4
                                      KNEERIGHT, ANKLERIGHT, ANKLERIGHT, // 5 6 7
                                      NECK, NECK, NECK, NECK, NECK, NECK, // 8 9 10 11 12 13
                                      HUMERUSLEFT, ELBOWLEFT, WRISTLEFT, WRISTLEFT,// 14 15 16 17
                                      HUMERUSRIGHT, ELBOWRIGHT, WRISTRIGHT, WRISTRIGHT}; // 18 19 20 21
  static final String[] jointToName_ = {"Left Hip", "Right Hip",
                                        "Left Knee", "Right Knee",
                                        "Left Ankle", "Right Ankle",
                                        "Left Collarbone", "Right Collarbone",
                                        "Left Shoulder", "Right Shoulder",
                                        "Left Elbow", "Right Elbow",
                                        "Left Wrist", "Right Wrist",
                                        "Spine", "Neck"};

  public static int jointToDof(int jnum) {
    return jointToDOF_[jnum];
  }
  public static int jointToLine(int jnum) {
    return jointToLine_[jnum];
  }
  public static String jointToName(int jnum) {
    return jointToName_[jnum];
  }
  public static int lineToJoint(int which, boolean isAend) {
    if (isAend) {
      return lineAToJoint_[which];
    } else {
      return lineBToJoint_[which];
    }
  }
  public SharedDouble jointValue(int jnum, int dof) {
    int side = jnum % 2;
    switch (jnum) {
    case HIPLEFT:
    case HIPRIGHT:
      switch (dof) {
      case 0: return hipforward_[side];
      case 1: return hipside_[side];
      case 2: return hiprotate_[side];
      default: return null;
      }
    case KNEELEFT:
    case KNEERIGHT:
      return knee_[side];
    case ANKLELEFT:
    case ANKLERIGHT:
      return ankle_[side];
    case SHOULDERLEFT:
    case SHOULDERRIGHT:
      switch (dof) {
      case 0: return shoulderforward_[side];
      case 1: return shoulderup_[side];
      default: return null;
      }
    case HUMERUSLEFT:
    case HUMERUSRIGHT:
      switch (dof) {
      case 0: return humerusforward_[side];
      case 1: return humerusside_[side];
      case 2: return humerusrotate_[side];
      default: return null;
      }
    case ELBOWLEFT:
    case ELBOWRIGHT:
      return elbow_[side];
    case WRISTLEFT:
    case WRISTRIGHT:
      switch (dof) {
      case 0: return wristup_[side];
      case 1: return wristside_[side];
      default: return null;
      }
    case SPINE:
      switch (dof) {
      case 0: return spineforward_;
      case 1: return spineside_;
      case 2: return spinerotate_;
      default: return null;
      }
    case NECK:
      switch (dof) {
      case 0: return neckforward_;
      case 1: return neckside_;
      case 2: return neckrotate_;
      default: return null;
      }
    default:
      return null;
    }
  }
  public static double jointMin(int jnum, int dof) {
    switch (jnum) {
    case HIPLEFT:
    case HIPRIGHT:
      switch (dof) {
      case 0: return -90*DEGREES; // hipforward
      case 1: return -30*DEGREES; // hipside
      case 2: return -45*DEGREES; // hiprotate
      default: return 0;
      }
    case KNEELEFT:
    case KNEERIGHT:
      return 0*DEGREES; // knee
    case ANKLELEFT:
    case ANKLERIGHT:
      return -15*DEGREES; // ankle
    case SHOULDERLEFT:
    case SHOULDERRIGHT:
      switch (dof) {
      case 0: return -30*DEGREES; // shoulderforward
      case 1: return 0*DEGREES; // shoulderup
      default: return 0;
      }
    case HUMERUSLEFT:
    case HUMERUSRIGHT:
      switch (dof) {
      case 0: return -90*DEGREES; // humerusforward
      case 1: return -45*DEGREES; // humerusside
      case 2: return -60*DEGREES; // humerusrotate
      default: return 0;
      }
    case ELBOWLEFT:
    case ELBOWRIGHT:
      return -10*DEGREES; // elbow
    case WRISTLEFT:
    case WRISTRIGHT:
      switch (dof) {
      case 0: return -90*DEGREES; // wristup
      case 1: return -60*DEGREES; // wristside
      default: return 0;
      }
    case SPINE:
      switch (dof) {
      case 0: return -30*DEGREES; // spineforward
      case 1: return -90*DEGREES; // spineside
      case 2: return -90*DEGREES; // spinerotate
      default: return 0;
      }
    case NECK:
      switch (dof) {
      case 0: return -75*DEGREES; // neckforward
      case 1: return -60*DEGREES; // neckside
      case 2: return -90*DEGREES; // neckrotate
      default: return 0;
      }
    default:
      return 0;
    }
  }
  public static double jointMax(int jnum, int dof) {
    switch (jnum) {
    case HIPLEFT:
    case HIPRIGHT:
      switch (dof) {
      case 0: return 180*DEGREES; // hipforward
      case 1: return 90*DEGREES; // hipside
      case 2: return 105*DEGREES; // hiprotate
      default: return 0;
      }
    case KNEELEFT:
    case KNEERIGHT:
      return 180*DEGREES; // knee
    case ANKLELEFT:
    case ANKLERIGHT:
      return 90*DEGREES; // ankle
    case SHOULDERLEFT:
    case SHOULDERRIGHT:
      switch (dof) {
      case 0: return 30*DEGREES; // shoulderforward
      case 1: return 45*DEGREES; // shoulderup
      default: return 0;
      }
    case HUMERUSLEFT:
    case HUMERUSRIGHT:
      switch (dof) {
      case 0: return 180*DEGREES; // humerusforward
      case 1: return 130*DEGREES; // humerusside
      case 2: return 180*DEGREES; // humerusrotate
      default: return 0;
      }
    case ELBOWLEFT:
    case ELBOWRIGHT:
      return 180*DEGREES; // elbow
    case WRISTLEFT:
    case WRISTRIGHT:
      switch (dof) {
      case 0: return 90*DEGREES; // wristup
      case 1: return 60*DEGREES; // wristside
      default: return 0;
      }
    case SPINE:
      switch (dof) {
      case 0: return 30*DEGREES; // spineforward
      case 1: return 90*DEGREES; // spineside
      case 2: return 90*DEGREES; // spinerotate
      default: return 0;
      }
    case NECK:
      switch (dof) {
      case 0: return 75*DEGREES; // neckforward
      case 1: return 60*DEGREES; // neckside
      case 2: return 90*DEGREES; // neckrotate
      default: return 0;
      }
    default:
      return 0;
    }
  }
  public void zeroAngles() {
    hipforward_[0].setDoubleValue(0);
    hipside_[0].setDoubleValue(0);
    hiprotate_[0].setDoubleValue(0);
    knee_[0].setDoubleValue(0);
    ankle_[0].setDoubleValue(0);
    hipforward_[1].setDoubleValue(0);
    hipside_[1].setDoubleValue(0);
    hiprotate_[1].setDoubleValue(0);
    knee_[1].setDoubleValue(0);
    ankle_[1].setDoubleValue(0);
    spineforward_.setDoubleValue(0);
    spineside_.setDoubleValue(0);
    spinerotate_.setDoubleValue(0);
    neckforward_.setDoubleValue(0);
    neckside_.setDoubleValue(0);
    neckrotate_.setDoubleValue(0);
    shoulderforward_[0].setDoubleValue(0);
    shoulderup_[0].setDoubleValue(0);
    humerusforward_[0].setDoubleValue(0);
    humerusside_[0].setDoubleValue(0);
    humerusrotate_[0].setDoubleValue(0);
    elbow_[0].setDoubleValue(0);
    wristup_[0].setDoubleValue(0);
    wristside_[0].setDoubleValue(0);
    shoulderforward_[1].setDoubleValue(0);
    shoulderup_[1].setDoubleValue(0);
    humerusforward_[1].setDoubleValue(0);
    humerusside_[1].setDoubleValue(0);
    humerusrotate_[1].setDoubleValue(0);
    elbow_[1].setDoubleValue(0);
    wristup_[1].setDoubleValue(0);
    wristside_[1].setDoubleValue(0);
  }
    
  
  Line3D[] getLines() {
    Line3D[] lines = new Line3D[lineCount_];
    int ii;
    for (ii=0; ii<lineCount_; ii++) {
      lines[ii] = new Line3D();
    }
    double length;
    
    // Lower spine and hips are fixed inverted-T shape.  The figure as a whole
    // needs to be rotated/positioned to adjust these
    ii = 0;
    lines[ii].a.x = 0; lines[ii].a.y = 0; lines[ii].a.z = 0;
    lines[ii].b.x = 0; lines[ii].b.y = 0; lines[ii].b.z = lowerspine_;
    ii++;
    lines[ii].a.x = -halfpelvis_; lines[ii].a.y=0; lines[ii].a.z = 0;
    lines[ii].b.x = halfpelvis_;  lines[ii].b.y=0; lines[ii].b.z = 0;

    // Legs
    for (int side=0; side<2; side++) {
      double sign = (side == 0) ? -1 : +1;
      // Femur:
      Matrix3D m1 = new Matrix3D(); m1.setYZrotate(-hipforward_[side].doubleValue());
      Matrix3D m2 = new Matrix3D(); m2.setXZrotate(-sign*hipside_[side].doubleValue());
      Matrix3D m3 = new Matrix3D(); m3.setXYrotate(-sign*hiprotate_[side].doubleValue());
      Matrix3D t1 = m1.multiply(m2);
      Matrix3D t2 = t1.multiply(m3);
      Point3D i = t2.multiply(new Point3D(1,0,0));
      Point3D j = t2.multiply(new Point3D(0,1,0));
      Point3D k = t2.multiply(new Point3D(0,0,1));
      ii++;
      lines[ii].a.x = sign*halfpelvis_; lines[ii].a.y = 0; lines[ii].a.z = 0;
      lines[ii].b.x = lines[ii].a.x - femur_ * k.x;  
      lines[ii].b.y = lines[ii].a.y - femur_ * k.y;  
      lines[ii].b.z = lines[ii].a.z - femur_ * k.z;
      
      // Tibia:
      // Need to rotate by knee_[s] angle in the YZ plane, but relative to the i,j,k
      // coordinate system
      Matrix3D m4 = new Matrix3D(); m4.setYZrotate(knee_[side].doubleValue());
      Point3D i2 = m4.multiply(new Point3D(1,0,0));
      Point3D j2 = m4.multiply(new Point3D(0,1,0));
      Point3D k2 = m4.multiply(new Point3D(0,0,1));
      // i2,j2,k2 are axes aligned with tibia, expressed in i,j,k coordinates.  Thus
      // in world coordinates they are i2.x*i+i2.y*j+i2.z*k etc
      Point3D i2world = new Point3D((i2.x * i.x + i2.y * j.x + i2.z * k.x),
                                    (i2.x * i.y + i2.y * j.y + i2.z * k.y),
                                    (i2.x * i.z + i2.y * j.z + i2.z * k.z));
      Point3D j2world = new Point3D((j2.x * i.x + j2.y * j.x + j2.z * k.x),
                                    (j2.x * i.y + j2.y * j.y + j2.z * k.y),
                                    (j2.x * i.z + j2.y * j.z + j2.z * k.z));
      Point3D k2world = new Point3D((k2.x * i.x + k2.y * j.x + k2.z * k.x),
                                    (k2.x * i.y + k2.y * j.y + k2.z * k.y),
                                    (k2.x * i.z + k2.y * j.z + k2.z * k.z));
      // Knee runs in the direction of -z in this frame
      ii++;
      lines[ii].a.x = lines[ii-1].b.x; lines[ii].a.y = lines[ii-1].b.y; lines[ii].a.z = lines[ii-1].b.z; 
      lines[ii].b.x = lines[ii].a.x - tibia_*k2world.x;
      lines[ii].b.y = lines[ii].a.y - tibia_*k2world.y;
      lines[ii].b.z = lines[ii].a.z - tibia_*k2world.z;
    
      // Foot:
      // Need to rotate by ankle_[s] in the YZ plane, but relative to the i2,j2,k2 coordinate system
      Matrix3D m5 = new Matrix3D();  m5.setYZrotate(ankle_[side].doubleValue());
        
      Point3D i3 = m5.multiply(new Point3D(1,0,0));
      Point3D j3 = m5.multiply(new Point3D(0,1,0));
      Point3D k3 = m5.multiply(new Point3D(0,0,1));
      // i3,j3,k3 are axes aligned with foot, expressed in i2,j2,k2 coordinates.  Thus
      // in world coordinates they are i3.x*i2world+i3.y*j2world+i3.z*k2world etc
      Point3D i3world = new Point3D((i3.x * i2world.x + i3.y * j2world.x + i3.z * k2world.x),
                                    (i3.x * i2world.y + i3.y * j2world.y + i3.z * k2world.y),
                                    (i3.x * i2world.z + i3.y * j2world.z + i3.z * k2world.z));
      Point3D j3world = new Point3D((j3.x * i2world.x + j3.y * j2world.x + j3.z * k2world.x),
                                    (j3.x * i2world.y + j3.y * j2world.y + j3.z * k2world.y),
                                    (j3.x * i2world.z + j3.y * j2world.z + j3.z * k2world.z));
      Point3D k3world = new Point3D((k3.x * i2world.x + k3.y * j2world.x + k3.z * k2world.x),
                                    (k3.x * i2world.y + k3.y * j2world.y + k3.z * k2world.y),
                                    (k3.x * i2world.z + k3.y * j2world.z + k3.z * k2world.z));
      // Foot runs in the direction of +y = j in this frame
      ii++;
      lines[ii].a.x = lines[ii-1].b.x; lines[ii].a.y = lines[ii-1].b.y; lines[ii].a.z = lines[ii-1].b.z; 
      lines[ii].b.x = lines[ii].a.x + foot_*j3world.x;
      lines[ii].b.y = lines[ii].a.y + foot_*j3world.y;
      lines[ii].b.z = lines[ii].a.z + foot_*j3world.z;
    }
    // Upper spine:
    // Start with spine in the positive z-direction; rotate in YZ plane by spineforward_ and 
    // in the XZ plane by spineside_.  
    Matrix3D m1 = new Matrix3D(); m1.setYZrotate(spineforward_.doubleValue());
    Matrix3D m2 = new Matrix3D(); m2.setXZrotate(spineside_.doubleValue());
    Matrix3D m2a = new Matrix3D(); m2a.setXYrotate(-spinerotate_.doubleValue());
    Matrix3D t1a = m1.multiply(m2);
    Matrix3D t1 = t1a.multiply(m2a);

    // Set up a dreibein that gets transformed as we move around the body    
    Point3D i = new Point3D(1,0,0);
    Point3D j = new Point3D(0,1,0);
    Point3D k = new Point3D(0,0,1);
    i = t1.multiply(i);
    j = t1.multiply(j);
    k = t1.multiply(k);
    // i,j,k is now a frame aligned with upper spine, expressed in world coords
    // Upper spine is in direction of new Z direction
    ii++;
    lines[ii].a.x = 0; lines[ii].a.y = 0; lines[ii].a.z = lowerspine_;
    lines[ii].b.x = lines[ii].a.x + upperspine_*k.x;
    lines[ii].b.y = lines[ii].a.y + upperspine_*k.y; 
    lines[ii].b.z = lines[ii].a.z + upperspine_*k.z;
    int spinetop = ii;

    {
      // Neck:
      // As with upper spine, rotate in (current) YZ plane by neckforward_ and in the (current) XZ 
      // plane by neckside_. Also rotate in the XY plane by neckrotate_
      Matrix3D m3 = new Matrix3D(); m3.setXYrotate(neckrotate_.doubleValue());
      Matrix3D m4 = new Matrix3D(); m4.setYZrotate(neckforward_.doubleValue());
      Matrix3D m5 = new Matrix3D(); m5.setXZrotate(neckside_.doubleValue());
      Matrix3D t2 = m4.multiply(m3);
      Matrix3D t3 = m5.multiply(t2);
      // Figure out new dreibein relative to i,j,k coord system. i,j,k in the i,j,k system
      // are obviously just (1,0,0) etc.
      Point3D i2 = t3.multiply(new Point3D(1,0,0));
      Point3D j2 = t3.multiply(new Point3D(0,1,0));
      Point3D k2 = t3.multiply(new Point3D(0,0,1));
      // k2 is direction of neck, expressed in terms of i, j, k 
      // (so in world coordinates, it is k2.x*i + k2.y*j + k2.z*k)
      Point3D i2world = new Point3D((i2.x*i.x + i2.y*j.x + i2.z*k.x),
                                    (i2.x*i.y + i2.y*j.y + i2.z*k.y),
                                    (i2.x*i.z + i2.y*j.z + i2.z*k.z));
      Point3D k2world = new Point3D((k2.x*i.x + k2.y*j.x + k2.z*k.x),
                                    (k2.x*i.y + k2.y*j.y + k2.z*k.y),
                                    (k2.x*i.z + k2.y*j.z + k2.z*k.z));
      ii++;
      lines[ii].a.x = lines[spinetop].b.x + EPSILON*k2world.x; 
      lines[ii].a.y = lines[spinetop].b.y + EPSILON*k2world.y; 
      lines[ii].a.z = lines[spinetop].b.z + EPSILON*k2world.z; 
      lines[ii].b.x = lines[ii].a.x + neck_*k2world.x;
      lines[ii].b.y = lines[ii].a.y + neck_*k2world.y; 
      lines[ii].b.z = lines[ii].a.z + neck_*k2world.z;
      // Head
      Point3D headbottom = new Point3D(lines[ii].b);
      Point3D headtop = new Point3D(headbottom);
      headtop.x += headheight_ * k2world.x;
      headtop.y += headheight_ * k2world.y;
      headtop.z += headheight_ * k2world.z;
      ii++;
      lines[ii].a.x = headbottom.x; lines[ii].a.y = headbottom.y; lines[ii].a.z = headbottom.z;
      lines[ii].b.x = lines[ii].a.x + (headheight_/2)*k2world.x + (headwidth_/2)*i2world.x;
      lines[ii].b.y = lines[ii].a.y + (headheight_/2)*k2world.y + (headwidth_/2)*i2world.y; 
      lines[ii].b.z = lines[ii].a.z + (headheight_/2)*k2world.z + (headwidth_/2)*i2world.z;
      ii++;
      lines[ii].a.x = lines[ii-1].b.x; lines[ii].a.y = lines[ii-1].b.y; lines[ii].a.z = lines[ii-1].b.z; 
      lines[ii].b.x = headtop.x;
      lines[ii].b.y = headtop.y;
      lines[ii].b.z = headtop.z;
      ii++;
      lines[ii].a.x = headbottom.x; lines[ii].a.y = headbottom.y; lines[ii].a.z = headbottom.z;
      lines[ii].b.x = lines[ii].a.x + (headheight_/2)*k2world.x - (headwidth_/2)*i2world.x;
      lines[ii].b.y = lines[ii].a.y + (headheight_/2)*k2world.y - (headwidth_/2)*i2world.y; 
      lines[ii].b.z = lines[ii].a.z + (headheight_/2)*k2world.z - (headwidth_/2)*i2world.z;
      ii++;
      lines[ii].a.x = lines[ii-1].b.x; lines[ii].a.y = lines[ii-1].b.y; lines[ii].a.z = lines[ii-1].b.z; 
      lines[ii].b.x = headtop.x;
      lines[ii].b.y = headtop.y;
      lines[ii].b.z = headtop.z;
    }
    // Shoulders and arms
    for (int side=0; side<2; side++) {
      double sign = (side == 0) ? -1 : +1;
      // At start, i,j,k are in world coordinates and are aligned with the upper spine

      // Shoulder:
      Matrix3D m6 = new Matrix3D(); m6.setXYrotate(-sign*shoulderforward_[side].doubleValue());
      Matrix3D m7 = new Matrix3D(); m7.setXZrotate(-sign*shoulderup_[side].doubleValue());
      Matrix3D t4 = m6.multiply(m7);
      Point3D i2 = t4.multiply(new Point3D(1,0,0));
      Point3D j2 = t4.multiply(new Point3D(0,1,0));
      Point3D k2 = t4.multiply(new Point3D(0,0,1));
      // i2,j2,k2 are axes aligned with shoulder, expressed in i,j,k coordinates.  Thus
      // in world coordinates they are i2.x*i+i2.y*j+i2.z*k etc
      Point3D i2world = new Point3D((i2.x * i.x + i2.y * j.x + i2.z * k.x),
                                    (i2.x * i.y + i2.y * j.y + i2.z * k.y),
                                    (i2.x * i.z + i2.y * j.z + i2.z * k.z));
      Point3D j2world = new Point3D((j2.x * i.x + j2.y * j.x + j2.z * k.x),
                                    (j2.x * i.y + j2.y * j.y + j2.z * k.y),
                                    (j2.x * i.z + j2.y * j.z + j2.z * k.z));
      Point3D k2world = new Point3D((k2.x * i.x + k2.y * j.x + k2.z * k.x),
                                    (k2.x * i.y + k2.y * j.y + k2.z * k.y),
                                    (k2.x * i.z + k2.y * j.z + k2.z * k.z));
      // Shoulder runs in (sign)i2 direction.  Note that we move the start points along the 
      // shoulder by a small amount so that it is possible to select them separately from
      // the neck joint
      ii++;
      lines[ii].a.x = lines[spinetop].b.x + sign*EPSILON*i2world.x; 
      lines[ii].a.y = lines[spinetop].b.y + sign*EPSILON*i2world.y; 
      lines[ii].a.z = lines[spinetop].b.z + sign*EPSILON*i2world.z; 
      lines[ii].b.x = lines[ii].a.x + sign*halfclavicle_*i2world.x;
      lines[ii].b.y = lines[ii].a.y + sign*halfclavicle_*i2world.y;
      lines[ii].b.z = lines[ii].a.z + sign*halfclavicle_*i2world.z;

      // Humerus
      Matrix3D m9 = new Matrix3D(); m9.setYZrotate(-humerusforward_[side].doubleValue());
      Matrix3D m10 = new Matrix3D(); m10.setXZrotate(-sign*humerusside_[side].doubleValue());
      Matrix3D m11 = new Matrix3D(); m11.setXYrotate(-sign*humerusrotate_[side].doubleValue());
      Matrix3D t5 = m9.multiply(m10);
      Matrix3D t6 = t5.multiply(m11);
      Point3D i3 = t6.multiply(new Point3D(1,0,0));
      Point3D j3 = t6.multiply(new Point3D(0,1,0));
      Point3D k3 = t6.multiply(new Point3D(0,0,1));
      // i3,j3,k3 are axes aligned with humerus, expressed in i2,j2,k2 coordinates.  Thus
      // in world coordinates they are i3.x*i2world+i3.y*j2world+i3.z*k2world etc
      Point3D i3world = new Point3D((i3.x * i2world.x + i3.y * j2world.x + i3.z * k2world.x),
                                    (i3.x * i2world.y + i3.y * j2world.y + i3.z * k2world.y),
                                    (i3.x * i2world.z + i3.y * j2world.z + i3.z * k2world.z));
      Point3D j3world = new Point3D((j3.x * i2world.x + j3.y * j2world.x + j3.z * k2world.x),
                                    (j3.x * i2world.y + j3.y * j2world.y + j3.z * k2world.y),
                                    (j3.x * i2world.z + j3.y * j2world.z + j3.z * k2world.z));
      Point3D k3world = new Point3D((k3.x * i2world.x + k3.y * j2world.x + k3.z * k2world.x),
                                    (k3.x * i2world.y + k3.y * j2world.y + k3.z * k2world.y),
                                    (k3.x * i2world.z + k3.y * j2world.z + k3.z * k2world.z));
      // Humerus runs in -k3 direction
      ii++;
      lines[ii].a.x = lines[ii-1].b.x; lines[ii].a.y = lines[ii-1].b.y; lines[ii].a.z = lines[ii-1].b.z; 
      lines[ii].b.x = lines[ii].a.x - humerus_*k3world.x;
      lines[ii].b.y = lines[ii].a.y - humerus_*k3world.y;
      lines[ii].b.z = lines[ii].a.z - humerus_*k3world.z;

      // Ulna
      Matrix3D m12 = new Matrix3D(); m12.setXZrotate(sign*elbow_[side].doubleValue());
      Point3D i4 = m12.multiply(new Point3D(1,0,0));
      Point3D j4 = m12.multiply(new Point3D(0,1,0));
      Point3D k4 = m12.multiply(new Point3D(0,0,1));
      // i4,j4,k4 are axes aligned with ulna, expressed in i3,j3,k3 coordinates.  Thus
      // in world coordinates they are i4.x*i3world+i4.y*j3world+i4.z*k3world etc
      Point3D i4world = new Point3D((i4.x * i3world.x + i4.y * j3world.x + i4.z * k3world.x),
                                    (i4.x * i3world.y + i4.y * j3world.y + i4.z * k3world.y),
                                    (i4.x * i3world.z + i4.y * j3world.z + i4.z * k3world.z));
      Point3D j4world = new Point3D((j4.x * i3world.x + j4.y * j3world.x + j4.z * k3world.x),
                                    (j4.x * i3world.y + j4.y * j3world.y + j4.z * k3world.y),
                                    (j4.x * i3world.z + j4.y * j3world.z + j4.z * k3world.z));
      Point3D k4world = new Point3D((k4.x * i3world.x + k4.y * j3world.x + k4.z * k3world.x),
                                    (k4.x * i3world.y + k4.y * j3world.y + k4.z * k3world.y),
                                    (k4.x * i3world.z + k4.y * j3world.z + k4.z * k3world.z));
      // Ulna runs in -k4 direction
      ii++;
      lines[ii].a.x = lines[ii-1].b.x; lines[ii].a.y = lines[ii-1].b.y; lines[ii].a.z = lines[ii-1].b.z; 
      lines[ii].b.x = lines[ii].a.x - ulna_*k4world.x;
      lines[ii].b.y = lines[ii].a.y - ulna_*k4world.y;
      lines[ii].b.z = lines[ii].a.z - ulna_*k4world.z;

      // Hand
      Matrix3D m13 = new Matrix3D(); m13.setXZrotate(sign*wristside_[side].doubleValue());
      Matrix3D m14 = new Matrix3D(); m14.setYZrotate(wristup_[side].doubleValue());
      Matrix3D t7 = m13.multiply(m14);
      Point3D i5 = t7.multiply(new Point3D(1,0,0));
      Point3D j5 = t7.multiply(new Point3D(0,1,0));
      Point3D k5 = t7.multiply(new Point3D(0,0,1));
      // i5,j5,k5 are axes aligned with ulna, expressed in i4,j4,k4 coordinates.  Thus
      // in world coordinates they are i5.x*i4world+i5.y*j4world+i5.z*k4world etc
      Point3D i5world = new Point3D((i5.x * i4world.x + i5.y * j4world.x + i5.z * k4world.x),
                                    (i5.x * i4world.y + i5.y * j4world.y + i5.z * k4world.y),
                                    (i5.x * i4world.z + i5.y * j4world.z + i5.z * k4world.z));
      Point3D j5world = new Point3D((j5.x * i4world.x + j5.y * j4world.x + j5.z * k4world.x),
                                    (j5.x * i4world.y + j5.y * j4world.y + j5.z * k4world.y),
                                    (j5.x * i4world.z + j5.y * j4world.z + j5.z * k4world.z));
      Point3D k5world = new Point3D((k5.x * i4world.x + k5.y * j4world.x + k5.z * k4world.x),
                                    (k5.x * i4world.y + k5.y * j4world.y + k5.z * k4world.y),
                                    (k5.x * i4world.z + k5.y * j4world.z + k5.z * k4world.z));
      ii++;
      lines[ii].a.x = lines[ii-1].b.x; lines[ii].a.y = lines[ii-1].b.y; lines[ii].a.z = lines[ii-1].b.z; 
      lines[ii].b.x = lines[ii].a.x - hand_*k5world.x;
      lines[ii].b.y = lines[ii].a.y - hand_*k5world.y;
      lines[ii].b.z = lines[ii].a.z - hand_*k5world.z;
    }

    return lines;
  }
  Line3D[] getBoundingBox() {
    Line3D[] lines = getLines();
    return getBoundingBox(lines);
  }
  final static Line3D[] getBoundingBox(Line3D[] lines) {
    int ii;
    Point3D min = new Point3D(Double.POSITIVE_INFINITY, Double.POSITIVE_INFINITY, Double.POSITIVE_INFINITY);
    Point3D max = new Point3D(Double.NEGATIVE_INFINITY, Double.NEGATIVE_INFINITY, Double.NEGATIVE_INFINITY);
    for (ii = 0; ii < lines.length; ii++) {
      if (lines[ii].a.x < min.x) min.x = lines[ii].a.x;
      if (lines[ii].b.x < min.x) min.x = lines[ii].b.x;
      if (lines[ii].a.y < min.y) min.y = lines[ii].a.y;
      if (lines[ii].b.y < min.y) min.y = lines[ii].b.y;
      if (lines[ii].a.z < min.z) min.z = lines[ii].a.z;
      if (lines[ii].b.z < min.z) min.z = lines[ii].b.z;
      if (lines[ii].a.x > max.x) max.x = lines[ii].a.x;
      if (lines[ii].b.x > max.x) max.x = lines[ii].b.x;
      if (lines[ii].a.y > max.y) max.y = lines[ii].a.y;
      if (lines[ii].b.y > max.y) max.y = lines[ii].b.y;
      if (lines[ii].a.z > max.z) max.z = lines[ii].a.z;
      if (lines[ii].b.z > max.z) max.z = lines[ii].b.z;
    }
    Line3D[] l = new Line3D[12];
    for (ii=0; ii<l.length; ii++) {
      l[ii] = new Line3D();
      l[ii].c = java.awt.Color.white;
    }
    ii=0;
    l[ii].a.x = min.x; l[ii].a.y = min.y; l[ii].a.z = min.z; l[ii].b.x = max.x; l[ii].b.y = min.y; l[ii].b.z = min.z; ii++;
    l[ii].a.x = max.x; l[ii].a.y = min.y; l[ii].a.z = min.z; l[ii].b.x = max.x; l[ii].b.y = max.y; l[ii].b.z = min.z; ii++;
    l[ii].a.x = max.x; l[ii].a.y = max.y; l[ii].a.z = min.z; l[ii].b.x = min.x; l[ii].b.y = max.y; l[ii].b.z = min.z; ii++;
    l[ii].a.x = min.x; l[ii].a.y = max.y; l[ii].a.z = min.z; l[ii].b.x = min.x; l[ii].b.y = min.y; l[ii].b.z = min.z; ii++;
    l[ii].a.x = min.x; l[ii].a.y = min.y; l[ii].a.z = max.z; l[ii].b.x = max.x; l[ii].b.y = min.y; l[ii].b.z = max.z; ii++;
    l[ii].a.x = max.x; l[ii].a.y = min.y; l[ii].a.z = max.z; l[ii].b.x = max.x; l[ii].b.y = max.y; l[ii].b.z = max.z; ii++;
    l[ii].a.x = max.x; l[ii].a.y = max.y; l[ii].a.z = max.z; l[ii].b.x = min.x; l[ii].b.y = max.y; l[ii].b.z = max.z; ii++;
    l[ii].a.x = min.x; l[ii].a.y = max.y; l[ii].a.z = max.z; l[ii].b.x = min.x; l[ii].b.y = min.y; l[ii].b.z = max.z; ii++;
    l[ii].a.x = min.x; l[ii].a.y = min.y; l[ii].a.z = min.z; l[ii].b.x = min.x; l[ii].b.y = min.y; l[ii].b.z = max.z; ii++;
    l[ii].a.x = max.x; l[ii].a.y = min.y; l[ii].a.z = min.z; l[ii].b.x = max.x; l[ii].b.y = min.y; l[ii].b.z = max.z; ii++;
    l[ii].a.x = max.x; l[ii].a.y = max.y; l[ii].a.z = min.z; l[ii].b.x = max.x; l[ii].b.y = max.y; l[ii].b.z = max.z; ii++;
    l[ii].a.x = min.x; l[ii].a.y = max.y; l[ii].a.z = min.z; l[ii].b.x = min.x; l[ii].b.y = max.y; l[ii].b.z = max.z; ii++;
    return l;
  }

  public String toString() {
    // Dump out all contents of the figure
    String result = "spine=" + Format.round(spineforward_) + " " + Format.round(spineside_) + " " + Format.round(spinerotate_) + " ";
    result += "neck=" + Format.round(neckforward_) + " " + Format.round(neckside_) + " " + Format.round(neckrotate_) + "\n";
    for (int ii=0; ii<2; ii++) {
      result += (ii == 0) ? "left {" : "right {";
      result += " hip=" + Format.round(hipforward_[ii]) + " " + Format.round(hipside_[ii]) + " " + Format.round(hiprotate_[ii]);
      result += " knee=" + Format.round(knee_[ii]);
      result += " ankle=" + Format.round(ankle_[ii]);
      result += " shoulder=" + Format.round(shoulderforward_[ii]) + " " + Format.round(shoulderup_[ii]);
      result += " humerus=" + Format.round(humerusforward_[ii]) + " " + Format.round(humerusside_[ii]) + " " + Format.round(humerusrotate_[ii]);
      result += " elbow=" + Format.round(elbow_[ii]);
      result += " wrist=" + Format.round(wristup_[ii]) + " " + Format.round(wristside_[ii]);
      result += " }\n";
    }
    return result;
  }
  public char[] fromString(char[] s) {
    s = Format.expect(s, "spine=");
    s = spineforward_.fromString(s);
    s = spineside_.fromString(s);
    s = spinerotate_.fromString(s);
    s = Format.expect(s, "neck=");
    s = neckforward_.fromString(s);
    s = neckside_.fromString(s);
    s = neckrotate_.fromString(s);
    for (int ii=0; ii<2; ii++) {
      s = Format.expect(s, (ii == 0) ? "left {" : "right {");
      s = Format.expect(s, "hip="); 
      s = hipforward_[ii].fromString(s);
      s = hipside_[ii].fromString(s);
      s = hiprotate_[ii].fromString(s);
      s = Format.expect(s, "knee=");
      s = knee_[ii].fromString(s);
      s = Format.expect(s, "ankle="); 
      s = ankle_[ii].fromString(s);
      s = Format.expect(s, "shoulder="); 
      s = shoulderforward_[ii].fromString(s); 
      s = shoulderup_[ii].fromString(s);
      s = Format.expect(s, "humerus=");
      s = humerusforward_[ii].fromString(s);
      s = humerusside_[ii].fromString(s);
      s = humerusrotate_[ii].fromString(s);
      s = Format.expect(s, "elbow=");
      s = elbow_[ii].fromString(s);
      s = Format.expect(s, "wrist=");
      s = wristup_[ii].fromString(s);
      s = wristside_[ii].fromString(s);
      s = Format.expect(s, "}");
    }
    return s;
  }
  public String fromString(String s) {
    char[] chars = s.toCharArray();
    return new String(fromString(chars));
  }
};
