// Copyright (c) 2006 David Drysdale.  See license.txt for license info
class FigurePosition extends Figure {
  // Position of mid-spine
  SharedDouble posx_;
  SharedDouble posy_;
  SharedDouble posz_;
  // Direction of normal to the inverted-T where pelvis meets lower spine
  SharedDouble ud_;
  SharedDouble lr_;
  SharedDouble tilt_;
  static boolean verbose_ = false;

  FigurePosition(Point3D pos, double ud, double lr, double tilt) {
    posx_ = new SharedDouble(pos.x);
    posy_ = new SharedDouble(pos.y);
    posz_ = new SharedDouble(pos.z);
    lr_ = new SharedDouble(lr);
    ud_ = new SharedDouble(ud);
    tilt_ = new SharedDouble(tilt);
  }
  FigurePosition(FigurePosition copy) {
    super(copy);
    posx_ = new SharedDouble(copy.posx_.doubleValue());
    posy_ = new SharedDouble(copy.posy_.doubleValue());
    posz_ = new SharedDouble(copy.posz_.doubleValue());
    lr_ = new SharedDouble(copy.lr_.doubleValue());
    ud_ = new SharedDouble(copy.ud_.doubleValue());
    tilt_ = new SharedDouble(copy.tilt_.doubleValue());
  }
  static final int POSITION = 16;
  
  public static int jointToDof(int jnum) {
    if (jnum==POSITION) return 3;
    else return Figure.jointToDof(jnum);
  }
  public static int jointToLine(int jnum) {
    if (jnum==POSITION) return 0; 
    else return Figure.jointToLine(jnum);
  }
  public static String jointToName(int jnum) {
    if (jnum==POSITION) return "Overall direction";
    else return Figure.jointToName(jnum);
  }
  public static int lineToJoint(int which, boolean isAend) {
    if ((which == 0) && isAend) return POSITION;
    else return Figure.lineToJoint(which, isAend);
  }
  public SharedDouble jointValue(int jnum, int dof) {
    if (jnum==POSITION) {
      switch (dof) {
      case 0: return lr_;
      case 1: return ud_;
      case 2: return tilt_;
      default: return null;
      }
    } else return super.jointValue(jnum, dof);
  }
  public static double jointMin(int jnum, int dof) {
    if (jnum==POSITION) {
      return -180*DEGREES;
    } else return Figure.jointMin(jnum, dof);
  }
  public static double jointMax(int jnum, int dof) {
    if (jnum==POSITION) {
      return 180*DEGREES;
    } else return Figure.jointMax(jnum, dof);
  }
  static public Matrix3D getTransformMatrix(double lr, double ud, double tilt) {
      Matrix3D m1 = new Matrix3D(); m1.setXYrotate(lr);
      Matrix3D m2 = new Matrix3D(); m2.setYZrotate(ud);
      Matrix3D m3 = new Matrix3D(); m3.setXZrotate(tilt);
      Matrix3D t1 = m1.multiply(m2);
      Matrix3D t = t1.multiply(m3);
      return t;
  }
  private void transform(Line3D[] lines) {
    // First rotate updown_ in the YZ plane, then rotate leftright_ in the XY plane
    for (int ii=0; ii<lines.length; ii++) {
      Matrix3D t = getTransformMatrix(lr_.doubleValue(), ud_.doubleValue(), tilt_.doubleValue());
      lines[ii].a = t.multiply(lines[ii].a);
      lines[ii].b = t.multiply(lines[ii].b);
    }
    // Translate
    if (verbose_) System.out.println("translate figure by (" + posx_.doubleValue() + ", " + posy_.doubleValue() + ", " + posz_.doubleValue() + ")");
    for (int ii=0; ii<lines.length; ii++) {
      lines[ii].a.x += posx_.doubleValue();
      lines[ii].b.x += posx_.doubleValue();
      lines[ii].a.y += posy_.doubleValue();
      lines[ii].b.y += posy_.doubleValue();
      lines[ii].a.z += posz_.doubleValue();
      lines[ii].b.z += posz_.doubleValue();
    }
  }
  public Line3D[] getBoundingBox() {
    Line3D[] lines = super.getLines();
    Line3D[] box = getBoundingBox(lines);
    transform(box);
    return box;
  }
  public Line3D[] getLines() {
    Line3D[] lines = super.getLines();
    transform(lines);
    return lines;
  }
  public void touchGround() {
    Line3D[] lines = getLines();
    double minz = Double.POSITIVE_INFINITY;
    for (int ii=0; ii<lines.length; ii++) {
      if (lines[ii].a.z < minz) minz = lines[ii].a.z;
      if (lines[ii].b.z < minz) minz = lines[ii].b.z;
    }
    posz_.setDoubleValue(posz_.doubleValue() - minz);
  }
  public String toString() {
    String result = "position=" + posx_.toString() + " " + posy_.toString() + " " + posz_.toString() + " ";
    result += ("direction=" +Format.round(lr_.doubleValue()) + " " + Format.round(ud_.doubleValue()) + " " 
               + Format.round(tilt_.doubleValue()) + "\n");
    result += super.toString();
    return result;
  }
  
  public char[] fromString(char[] s) {
    s = Format.expect(s, "position=");
    s = posx_.fromString(s);
    s = posy_.fromString(s);
    s = posz_.fromString(s);
    s = Format.expect(s, "direction=");
    s = lr_.fromString(s);
    s = ud_.fromString(s);
    s = tilt_.fromString(s);
    s = super.fromString(s);
    return s;
  }
  public String fromString(String s) {
    char[] chars = s.toCharArray();
    return new String(fromString(chars));
  }
};
