// Copyright (c) 2006 David Drysdale.  See license.txt for license info
class Matrix3D {
  public double a11;
  public double a12;
  public double a13;
  public double a21;
  public double a22;
  public double a23;
  public double a31;
  public double a32;
  public double a33;
  public Matrix3D() {}
  public Matrix3D(Matrix3D copy) {
    a11 = copy.a11; a12 = copy.a12; a13 = copy.a13;
    a21 = copy.a21; a22 = copy.a22; a23 = copy.a23;
    a31 = copy.a31; a32 = copy.a32; a33 = copy.a33;
  }
  public void setYZrotate(double a) {
    a11 = 1; a12 = 0;            a13 = 0;       
    a21 = 0; a22 = Math.cos(a);  a23 = Math.sin(a);
    a31 = 0; a32 = -Math.sin(a); a33 = Math.cos(a);
  }
  public void setXZrotate(double a) {
    a11 =  Math.cos(a); a12 = 0; a13 = Math.sin(a);
    a21 = 0;            a22 = 1; a23 = 0;
    a31 = -Math.sin(a); a32 = 0; a33 = Math.cos(a);
  }
  public void setXYrotate(double a) {
    a11 = Math.cos(a); a12 = -Math.sin(a); a13 = 0;
    a21 = Math.sin(a); a22 =  Math.cos(a); a23 = 0;
    a31 = 0;           a32 = 0;            a33 = 1;  
  }
  public Point3D multiply(Point3D pt) {
    Point3D result = new Point3D();
    result.x = a11 * pt.x + a12 * pt.y + a13 * pt.z;
    result.y = a21 * pt.x + a22 * pt.y + a23 * pt.z;
    result.z = a31 * pt.x + a32 * pt.y + a33 * pt.z;
    return result;
  }
  public Matrix3D multiply(Matrix3D m) {
    Matrix3D result = new Matrix3D();
    result.a11 = a11 * m.a11 + a12 * m.a21 + a13 * m.a31;
    result.a12 = a11 * m.a12 + a12 * m.a22 + a13 * m.a32;
    result.a13 = a11 * m.a13 + a12 * m.a23 + a13 * m.a33;
    result.a21 = a21 * m.a11 + a22 * m.a21 + a23 * m.a31;
    result.a22 = a21 * m.a12 + a22 * m.a22 + a23 * m.a32;
    result.a23 = a21 * m.a13 + a22 * m.a23 + a23 * m.a33;
    result.a31 = a31 * m.a11 + a32 * m.a21 + a33 * m.a31;
    result.a32 = a31 * m.a12 + a32 * m.a22 + a33 * m.a32;
    result.a33 = a31 * m.a13 + a32 * m.a23 + a33 * m.a33;
    return result;
  }
  public void dump(String prefix) {
    System.out.println(prefix + "(" + a11 + ", " + a12 + ", "+ a13 + ")");
    System.out.println(prefix + "(" + a21 + ", " + a22 + ", "+ a23 + ")");
    System.out.println(prefix + "(" + a31 + ", " + a32 + ", "+ a33 + ")");
  }
};