// Copyright (c) 2006 David Drysdale.  See license.txt for license info
class Move {
  static java.util.Hashtable sTable_ = new java.util.Hashtable();
  String name_;
  static boolean verbose_ = false;

  Move(String name) {
    // No double quotes allowed in names
    name_ = name.replace('\"', '\'');
  }
  public FiguresSequence getFigSeq() {
    return null;
  }

  static public Move lookup(String name) {
    Object o = sTable_.get(name);
    if (o instanceof Move) {
      return (Move)o;
    } else {
      return null;
    }
  }
  static public void register(String name, Move move) {
    sTable_.put(name, move);
  }
  public String toString() {
    return new String("move=\""+name_+"\"");
  }
  public char[] fromString(char[] s) {
    s = Format.expect(s, "move=\"");
    // look for the closing double quote
    int jj = 0;
    while (s[jj] != '\"') jj++;
    name_ = new String(s, 0, jj);
    jj++;
    // Return what's left of the string
    char[] left = new char[s.length - jj];
    System.arraycopy(s, jj, left, 0, s.length-jj);
    return left;
  }
  public String fromString(String s) {
    char[] chars = s.toCharArray();
    return new String(fromString(chars));
  }
  static public String dump() {
    java.util.Enumeration e = sTable_.keys();
    String result = new String();
    while (e.hasMoreElements()) {
      Object key = e.nextElement();
      if (key instanceof String) {
        String name = (String)key;
        Move move = lookup(name);
        result += move;
      }
    }
    return result;
  }
  static public void load(String sdata) {
    char[] s = sdata.toCharArray();
    try {
      while (true) {
        // This position is the beginning of the move description.  However, we need to
        // scan forward first to figure out what type of move it is.
        Move m = new Move("");
        char[] t = m.fromString(s);
        
        if (verbose_) System.out.println("next up is a move called " + m.name_);
        t = Format.expect(t, "type=");
        boolean isSimple = false;
        try {
          Format.expect(t,"simple");
          isSimple = true;
        } catch (NumberFormatException e) {}
        boolean isCompound = false;
        try {
          Format.expect(t,"compound");
          isCompound = true;
        } catch (NumberFormatException e) {}
        if (isSimple) {
          SimpleMove smove = new SimpleMove(m.name_, new FiguresSequence(1,1));
          s = smove.fromString(s);
          if (verbose_) System.out.println("loaded simple move called "+smove.name_);
        } else if (isCompound) {
          CompoundMove cmove = new CompoundMove(m.name_, new String[0]);
          s = cmove.fromString(s);
          if (verbose_) System.out.println("loaded compound move "+cmove);
        } else {
          throw new NumberFormatException("unexpected move type in "+new String(t));
        }
      }
    } catch (Exception e) {
      if (verbose_) System.out.println("loop terminated by exception " + e);
    }
  }
};