// Copyright (c) 2006 David Drysdale.  See license.txt for license info
class Point3D {
  public double x;
  public double y;
  public double z;
  public Point3D() {
    x = y = z = 0.0;
  }
  public Point3D(Point3D copy) {
    x = copy.x; y = copy.y; z = copy.z;
  }
  public Point3D(double a, double b, double c) {
    x = a; y = b; z = c;
  }
  public Point3D(String s) {
    fromString(s);
  }
  public double length() {
    return Math.sqrt(x*x + y*y + z*z);
  }
  public Point3D vectorProduct(Point3D other) {
    Point3D result = new Point3D();
    result.x = y*other.z - z*other.y;
    result.y = z*other.x - x*other.z;
    result.z = x*other.y - y*other.x;
    return result;
  }
  public void makeUnit() {
    double len = length();
    x /= len;
    y /= len;
    z /= len;
  }
  public String toString() {
    return "(" + x + ", " + y + ", " + z + ")";
  }
  public char[] fromString(char[] s) {
    // Parse format as above
    int ii=0;
    while (s[ii] == ' ') ii++;
    if (s[ii] != '(') throw new NumberFormatException("expecting (");
    ii++;
    int jj=ii;
    // Find the comma
    while (s[jj] != ',') jj++;
    // x is everything between ii and jj (first nonblank and comma)
    String xs = new String(s, ii, jj-ii);
    // Move past comma and spaces
    ii=jj+1;
    while (s[ii] == ' ') ii++;
    // Find the comma
    jj=ii;
    while (s[jj] != ',') jj++;
    String ys = new String(s, ii, jj-ii);
    // Move past comma and spaces
    ii=jj+1;
    while (s[ii] == ' ') ii++;
    // Find the end parenthesis
    jj=ii;
    while (s[jj] != ')') jj++;
    String zs = new String(s, ii, jj-ii);
    jj++;

    Double xd = new Double(xs); 
    Double yd = new Double(ys); 
    Double zd = new Double(zs); 
    x = xd.doubleValue();
    y = yd.doubleValue();
    z = zd.doubleValue();
    
    // Return what's left of the string
    char[] left = new char[s.length - jj];
    System.arraycopy(s, jj, left, 0, s.length -jj);
    return left;
  }
  public String fromString(String s) {
    char[] chars = s.toCharArray();
    return new String(fromString(chars));
  }
  public void dump(String prefix) {
    System.out.println(prefix + toString());
  }
};
