// Copyright (c) 2006 David Drysdale.  See license.txt for license info
class Sequence {
  // deltat_[0] is used as offset from a previous sequence
  public double[] deltat_;
  Sequence(int numFrames) {
    deltat_ = new double[numFrames];
  }
  Sequence(Sequence other) {
    deltat_ = new double[other.length()];
    for (int ii=0; ii<other.length(); ii++) {
      deltat_[ii] = other.deltat_[ii];
    }
  }
  public void addFrame(int frame) {
    int numFrames = deltat_.length;
    if (frame>=numFrames) return;
    double[] newdt = new double[numFrames+1];
    int ii = 0;
    int jj = 0;
    while (ii<numFrames) {
      newdt[jj] = deltat_[ii];
      jj++;
      if (ii == frame) {
        // insert an extra
        newdt[jj] = deltat_[ii];
        jj++;
      } 
      ii++;
    }
    deltat_ = newdt;
  }
  public void delFrame(int frame) {
    int numFrames = deltat_.length;
    if (frame>=numFrames) return;
    double[] newdt = new double[numFrames-1];
    int ii = 0;
    int jj = 0;
    while (ii<numFrames) {
      if (ii != frame) {
        newdt[jj] = deltat_[ii];
        jj++;
      }
      ii++;
    }
    deltat_ = newdt;
  }
  public int length() {
    return deltat_.length;
  }
  public double totalDuration() {
    double sum = 0.0;
    for (int ii=0; ii<deltat_.length; ii++) {
      sum += deltat_[ii];
    }
    return sum;
  }
  public int nearestFrame(double t) {
    double cumulative = 0.0;
    for (int ii=0; ii<deltat_.length; ii++) {
      double next = cumulative;
      if (ii == (deltat_.length-1)) {
        next += deltat_[0];
      } else {
        next += deltat_[ii+1];
      }
      // Range from [ii, ii+1) is time [cumulative, next)
      if ((t >= cumulative) && (t < next)) {
        if ((t-cumulative) > (next-t)) {
          if (ii >= (deltat_.length-1)) {
            return 0;
          } else {
            return ii+1;
          }
        } else {
          return ii;
        }
      }
      cumulative = next;
    }
    return -1;
  }
  public double getTime(int which) {
    double sum = 0.0;
    for (int ii=0; ii<which; ii++) {
      sum += deltat_[ii];
    }
    return sum;
  }
  public String toString() {
    String result = "frames[" + deltat_.length + "]=";
    for (int ii=0; ii<deltat_.length; ii++) {
      result += deltat_[ii] + " ";
    }
    return result;
  }
  public char[] fromString(char[] s) {
    int ii=0;
    s = Format.expect(s, "frames[");
    // look for the ']'
    int jj = ii;
    while (s[jj] != ']') jj++;
    String ns = new String(s, ii, jj-ii);
    int numFrames = Integer.parseInt(ns, 10);
    deltat_ = new double[numFrames];
    if (s[jj+1] != '=') throw new NumberFormatException("expected '='");
    ii = jj+2; // move past ']='
    for (int frame=0; frame<numFrames; frame++) {
      // Skip leading whitespace
      while (Format.isWhitespace(s[ii])) ii++;
      // Find first non-number character
      jj=ii;
      while  ((jj<s.length) && (SharedDouble.isNumChar(s[jj]))) jj++;
      String ds = new String(s, ii, jj-ii);
      Double dd = new Double(ds);
      deltat_[frame] = dd.doubleValue();
      ii = jj;
    }

    // Return what's left of the string
    char[] left = new char[s.length - jj];
    System.arraycopy(s, jj, left, 0, s.length -jj);
    return left;
  }
  public String fromString(String s) {
    char[] chars = s.toCharArray();
    return new String(fromString(chars));
  }
};