// Copyright (c) 2006 David Drysdale.  See license.txt for license info
import java.applet.*;
import java.awt.*;


public class TwigAnim extends Applet implements Triggerable {
  Component theFrame_;

  // Top row of widgets
  Label jointLabel_;
  TextField moveName_;
  Button editButton_;
  Button loadButton_;
  Button saveButton_;
  Button addButton_;
  Button delButton_;
  Label frameLabel_;
  // Second row of widgets
  double speed_ = 2;
  TextField speedField_;
  Checkbox repeatCheckbox_;
  Button goButton_;
  Button stopButton_;
  Button resetButton_;
  Label timeLabel_;
  // Bottom row: single canvas
  AnimCanvas plot_;

  boolean editMode_ = false;
  int editFrame_;

  Animator animator_;
  // Thread to flow time, and an indicator of whether it is running
  AnimateThread animThread_;
  boolean running_ = false;
  SharedDouble curt_;
  // Camera object
  Camera camera_;
  // Sequence of data to display
  Line3DSequence frames3d_;
  FiguresSequence figures_;
  boolean includeBB_ = false;
  boolean includeAxes_ = true;
  final static String startmove = "quartet move";

  public String getAppletInfo() {
    return "TwigAnim (c) David Drysdale 2006";
  }

  void setAxes(double I, Line3D[] l, int offset) {
    int ii = offset;
    l[ii].a.x = 0; l[ii].a.y = 0; l[ii].a.z = 0; l[ii].b.x = 0; l[ii].b.y = 0; l[ii].b.z = I; l[ii].c = Color.red; ii++;
    l[ii].a.x = 0; l[ii].a.y = 0; l[ii].a.z = 0; l[ii].b.x = 0; l[ii].b.y = I; l[ii].b.z = 0; l[ii].c = Color.white; ii++;
    l[ii].a.x = 0; l[ii].a.y = 0; l[ii].a.z = 0; l[ii].b.x = I; l[ii].b.y = 0; l[ii].b.z = 0; l[ii].c = Color.blue; ii++;
  }

  void setFigure(FigurePosition figure, int frame, int offset) {
    figure.touchGround();
    Line3D[] figlines = figure.getLines();
    for (int jj = 0; jj < figlines.length; jj++) {
      frames3d_.lines_[frame][offset+jj] = figlines[jj];
    }
    if (includeBB_) {
      Line3D[] figbox = figure.getBoundingBox();
      for (int jj = 0; jj < figbox.length; jj++) {
        frames3d_.lines_[frame][offset+Figure.lineCount_+jj] = figbox[jj];
      }
    }
  }

  // Applet initialization
  public void init() {
    // find an owning frame
    theFrame_ = getParent();
    while ((theFrame_ != null) && !(theFrame_ instanceof Frame)) {
      theFrame_ = theFrame_.getParent();
    }
    
    // Set up the applet layout
    Dimension d = new Dimension();
    d = size();
    Panel entryPanel = new Panel();
    jointLabel_ = new Label("                                           ");
    entryPanel.add(jointLabel_);
    moveName_ = new TextField(startmove);
    entryPanel.add(moveName_);
    loadButton_ = new Button("Load");
    loadButton_.enable();
    entryPanel.add(loadButton_);
    saveButton_ = new Button("Save");
    saveButton_.enable();
    entryPanel.add(saveButton_);
    editButton_ = new Button("Edit");
    editButton_.enable();
    entryPanel.add(editButton_);
    addButton_ = new Button("Add");
    addButton_.disable();
    entryPanel.add(addButton_);
    delButton_ = new Button("Del");
    delButton_.disable();
    entryPanel.add(delButton_);
    frameLabel_ = new Label("                           ");
    entryPanel.add(frameLabel_);
    this.add(entryPanel);

    entryPanel = new Panel();
    entryPanel.add(new Label("Speed:"));
    speedField_ = new TextField("    " );
    speedField_.setText("" + speed_);
    entryPanel.add(speedField_);
    repeatCheckbox_ = new Checkbox("Loop");
    repeatCheckbox_.setState(true);
    entryPanel.add(repeatCheckbox_);
    goButton_ = new Button("Go");
    entryPanel.add(goButton_);
    goButton_.enable();
    stopButton_ = new Button("Stop");
    entryPanel.add(stopButton_);
    stopButton_.disable();
    resetButton_ = new Button("Reset");
    resetButton_.enable();
    entryPanel.add(resetButton_);
    timeLabel_ = new Label("                     ");
    entryPanel.add(timeLabel_);
    this.add(entryPanel);

    // Set up the master world-clock
    curt_ = new SharedDouble(0.0);

    // Camera
    camera_ = new Camera(this, new Point3D(4,-4,4), 2);

    // Canvas
    plot_ = new AnimCanvas(-1.5,-1.0, 1.5,1.0, camera_);
    plot_.setSize(d.width, d.height - entryPanel.preferredSize().height - 10); 
    this.add(plot_, -1);
    
    // Generate a FigureSequence
    figures_ = new FiguresSequence(1,1);
    figures_.fromString(savedPositions_);
    generateFrames();

    // Load up the move database
    Move.load(savedMoves_);
    Move move = Move.lookup(startmove);
    figures_ = move.getFigSeq();
    generateFrames();

    // Point the camera in the right direction
    camera_.setDir(frames3d_); 
  }
  public void generateFrames() {
    // Generate 3D frame sequence from the FiguresSequence
    frames3d_ = figures_.getLines3D(includeBB_, includeAxes_);
    // Animator of frame sequence
    animator_ = new Animator(plot_, timeLabel_, frames3d_);
    // Display of frame info
    if (editMode_) {
      if (figures_.length() > 2) { 
        delButton_.enable();
      } else {
        delButton_.disable();
      }
      frameLabel_.setText("Frame = "+(editFrame_+1) + " of "+figures_.length());
    }
  }
  public boolean action(Event e, Object what) {
    boolean processed = false;
    if (e.target == goButton_) {
      goButton_.disable();
      resetButton_.disable();
      stopButton_.enable();
      running_ = true;
      animThread_ = new AnimateThread(this, repeatCheckbox_, 
                                      animator_.totalDuration(), animator_.activeDuration(true),  
                                      curt_); 
      animThread_.setPriority(Thread.MIN_PRIORITY);
      setSpeed();
      animThread_.start();
    } else if (e.target == stopButton_) {
      if (running_) {
        animThreadDone();
      }
    } else if (e.target == resetButton_) {
      curt_.setDoubleValue(0.0);
      trigger();
    } else if (e.target == speedField_) {
      setSpeed();
    } else if (e.target == editButton_) {
      if (editMode_) {
        leaveEditMode();
      } else {
        enterEditMode();
        // Set up a joint selection
        plot_.setClickConsumer(new JointSelector(this, plot_, camera_, jointLabel_, 
                                                 figures_.figure[editFrame_], figures_.color));
      }
    } else if ((e.target == addButton_) && editMode_) {
      figures_.addFrame(editFrame_);
      generateFrames();
      trigger();
    } else if ((e.target == delButton_) && editMode_) {
      figures_.delFrame(editFrame_);
      if (editFrame_ >= figures_.length()) editFrame_ = figures_.length() - 1;
      // Change the joint selection process to allow for a different frame
      plot_.setControl(null);
      plot_.setDragConsumer(camera_);
      plot_.setClickConsumer(new JointSelector(this, plot_, camera_, jointLabel_, 
                                               figures_.figure[editFrame_], figures_.color));
      generateFrames();
      trigger();
    } else if (e.target == saveButton_) {
      SaveDialog save = new SaveDialog((Frame)theFrame_, true, Move.dump());
      save.resize(800, 400);
      Dimension d = size();
      save.move((d.width/2) - (save.size().width/2), 
                (d.height/2) + 100 - (save.size().height/2));
      ((Window)save).show();
      save.dispose();
    } else if (e.target == loadButton_) {
      LoadDialog load = new LoadDialog((Frame)theFrame_, true);
      load.resize(800, 400);
      Dimension d = size();
      load.move((d.width/2) - (load.size().width/2), 
                (d.height/2) + 100 - (load.size().height/2));
      // Note that we explicitly use the show() method of Window, not
      // Dialog.  This allows the code to work on JDK 1.02, since show()
      // only became a method of Dialog in JDK 1.1 (passim)
      ((Window)load).show();
      load.dispose();
      if (load.getAnswer()) {
        String newValue = load.getText();
        try {
          Move.load(newValue);
          setMove();
        } catch (NumberFormatException fe) {
          ErrorDialog errdialog = new ErrorDialog((Frame)theFrame_, fe.toString());
          errdialog.show();
        }
      }
    } else if (e.target == moveName_) {
      setMove();
    }
    processed = true;
    return(processed);
  }
  private void leaveEditMode() {
    editMode_ = false;
    stopButton_.disable();
    goButton_.enable();
    resetButton_.enable();
    editButton_.setLabel("Edit");
    jointLabel_.setText("");
    // Turn off all the added extras
    plot_.setClickConsumer(null);
    plot_.setDragConsumer(camera_);
    plot_.setControl(null);
    addButton_.disable();
    delButton_.disable();
    frameLabel_.setText("");
    // May have changed the centre of the tableau so repoint the camera
    camera_.setDir(frames3d_); 
    trigger();
  }
  private void enterEditMode() {
    editMode_ = true;
    // Turn everything off
    animThreadDone();
    goButton_.disable();
    resetButton_.disable();
    stopButton_.disable();
    
    // Change the edit button
    editButton_.setLabel("Show");
    editFrame_ = frames3d_.nearestFrame(curt_.doubleValue());
    curt_.setDoubleValue(frames3d_.getTime(editFrame_));
    trigger();
    frameLabel_.setText("Frame = "+(editFrame_+1) + " of "+figures_.length());
    addButton_.enable();
    if (figures_.length() > 2) delButton_.enable();
  }
  private void setMove() {
    String movename = moveName_.getText();
    Move move = Move.lookup(movename);
    if (move == null) {
      ErrorDialog errdialog = new ErrorDialog((Frame)theFrame_, "Failed to find move \""+movename+"\"");
      errdialog.show();
    } else {
      figures_ = move.getFigSeq();
      curt_.setDoubleValue(0.0);
      leaveEditMode();
      editFrame_ = 0;
      generateFrames();
      camera_.setDir(frames3d_);
      trigger();
    }
  }
  private void setSpeed() {
    String value = speedField_.getText();
    Double speed = new Double(value);
    if (animThread_ != null) animThread_.setSpeed(speed.doubleValue());
  }
  public void animThreadDone() {
    if (running_) {
      stopButton_.disable();
      goButton_.enable();
      resetButton_.enable();
      running_ = false;
      animThread_.term();
    }
  }
  public void trigger() {
    // We have been triggered because something has changed -- either time has moved on
    // (the AnimateThread object triggers) or something has changed (the camera, the model etc)
    if (editMode_) {
      animator_.drawFrame(editFrame_, curt_.doubleValue(), repeatCheckbox_.getState(), camera_);
    } else {
      animator_.drawFrame(curt_.doubleValue(), repeatCheckbox_.getState(), camera_);
    }
  }
  final String savedPositions_ =   
    "frames[10]=2.0 2.0 2.0 2.0 2.0 2.0 2.0 2.0 2.0 2.0 \n"+
    "figures[2]=\n"+
    "figure 0 color=ff00 {\n"+
    "frame 0 {\n"+
    "position=0.5 0.5 0.92 direction=0.0 0.0 0.0\n"+
    "spine=0.0 0.0 0.0 neck=0.0 0.0 0.0\n"+
    "left { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "right { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "}\n"+
    "frame 1 {\n"+
    "position=0.5 0.5 0.92 direction=0.0 0.0 0.0\n"+
    "spine=0.261 0.13 0.0 neck=0.785 0.523 0.0\n"+
    "left { hip=0.785 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "right { hip=0.0 0.785 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "}\n"+
    "frame 2 {\n"+
    "position=0.5 0.5 0.92 direction=0.0 0.0 0.0\n"+
    "spine=0.0 0.0 0.0 neck=0.785 0.523 0.0\n"+
    "left { hip=0.785 0.0 0.785 knee=0.785 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "right { hip=0.0 0.785 0.0 knee=0.785 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "}\n"+
    "frame 3 {\n"+
    "position=0.5 0.5 0.92 direction=0.0 0.0 0.0\n"+
    "spine=0.0 0.0 0.0 neck=0.0 0.0 0.785\n"+
    "left { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "right { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "}\n"+
    "frame 4 {\n"+
    "position=0.5 0.5 0.92 direction=0.0 0.0 0.0\n"+
    "spine=0.0 0.0 0.0 neck=0.0 0.0 -0.785\n"+
    "left { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.392 0.0 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "right { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.392 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "}\n"+
    "frame 5 {\n"+
    "position=0.5 0.5 0.92 direction=0.0 0.0 0.0\n"+
    "spine=0.0 0.0 0.0 neck=0.0 0.0 0.0\n"+
    "left { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.523 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "right { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.392 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "}\n"+
    "frame 6 {\n"+
    "position=0.5 0.5 0.92 direction=0.0 0.0 0.0\n"+
    "spine=0.0 0.0 0.0 neck=0.0 0.0 0.0\n"+
    "left { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.523 0.0 0.0 elbow=1.047 wrist=0.0 0.0 }\n"+
    "right { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.392 0.0 elbow=0.523 wrist=0.0 0.0 }\n"+
    "}\n"+
    "frame 7 {\n"+
    "position=0.5 0.5 0.92 direction=0.0 0.0 0.0\n"+
    "spine=0.0 0.0 0.0 neck=0.0 0.0 0.0\n"+
    "left { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.523 0.0 0.392 elbow=1.047 wrist=0.0 0.0 }\n"+
    "right { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.392 0.0 elbow=0.523 wrist=0.0 0.0 }\n"+
    "}\n"+
    "frame 8 {\n"+
    "position=0.5 0.5 0.92 direction=0.0 0.0 0.0\n"+
    "spine=0.0 0.0 0.0 neck=0.0 0.0 0.0\n"+
    "left { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.523 0.0 0.392 elbow=1.047 wrist=0.392 0.0 }\n"+
    "right { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.392 0.0 elbow=0.523 wrist=0.0 0.392 }\n"+
    "}\n"+
    "frame 9 {\n"+
    "position=0.5 0.5 0.92 direction=0.0 0.0 0.0\n"+
    "spine=0.0 0.0 0.0 neck=0.0 0.0 0.0\n"+
    "left { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.523 0.0 0.392 elbow=1.047 wrist=0.392 0.392 }\n"+
    "right { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.392 0.0 elbow=0.523 wrist=0.392 0.392 }\n"+
    "}\n"+
    "}\n"+
    "figure 1 color=ff0000 {\n"+
    "frame 0 {\n"+
    "position=0.5 -0.5 0.92 direction=3.0 0.0 0.0\n"+
    "spine=0.0 0.0 0.0 neck=0.0 0.0 0.0\n"+
    "left { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "right { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "}\n"+
    "frame 1 {\n"+
    "position=0.5 -0.5 0.92 direction=2.5 0.0 0.0\n"+
    "spine=0.0 0.0 0.0 neck=0.0 0.0 0.0\n"+
    "left { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "right { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "}\n"+
    "frame 2 {\n"+
    "position=0.5 -0.5 0.92 direction=2.0 0.0 0.0\n"+
    "spine=0.0 0.0 0.0 neck=0.0 0.0 0.0\n"+
    "left { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "right { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "}\n"+
    "frame 3 {\n"+
    "position=0.5 -0.5 0.92 direction=1.5 0.0 0.0\n"+
    "spine=0.0 0.0 0.0 neck=0.0 0.0 0.0\n"+
    "left { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "right { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "}\n"+
    "frame 4 {\n"+
    "position=0.5 -0.5 0.92 direction=1.5 0.5 0.0\n"+
    "spine=0.0 0.0 0.0 neck=0.0 0.0 0.0\n"+
    "left { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "right { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "}\n"+
    "frame 5 {\n"+
    "position=0.5 -0.5 0.92 direction=1.5 1.0 0.0\n"+
    "spine=0.0 0.0 0.0 neck=0.0 0.0 0.0\n"+
    "left { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "right { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "}\n"+
    "frame 6 {\n"+
    "position=0.5 -0.5 0.92 direction=1.5 0.5 0.0\n"+
    "spine=0.0 0.0 0.0 neck=0.0 0.0 0.0\n"+
    "left { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "right { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "}\n"+
    "frame 7 {\n"+
    "position=0.5 -0.5 0.92 direction=1.5 0.0 0.0\n"+
    "spine=0.0 0.0 0.0 neck=0.0 0.0 0.0\n"+
    "left { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "right { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "}\n"+
    "frame 8 {\n"+
    "position=0.5 -0.5 0.92 direction=2.0 0.0 0.0\n"+
    "spine=0.0 0.0 0.0 neck=0.0 0.0 0.0\n"+
    "left { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "right { hip=0.0 0.0 0.0 knee=0.3 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "}\n"+
    "frame 9 {\n"+
    "position=0.5 -0.5 0.92 direction=2.5 0.0 0.0\n"+
    "spine=0.0 0.0 0.0 neck=0.0 0.0 0.0\n"+
    "left { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "right { hip=0.0 0.0 0.0 knee=0.3 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.0 0.0 elbow=0.0 wrist=0.0 0.0 }\n"+
    "}\n"+
    "}\n";
  final String savedMoves_ =
    "move=\"parts move\" type=compound contents[4]=\"test move|test move red@(1,0,0)?(1.5708,0,0)\" + " + 
                                                  "\"test move|test move red?(1.5708,0,0)@(1,0,0)\" + " +
                                                  "\"test move|test move red?(1.5708,0,0)@(1,0,0)\" + " +
                                                  "\"test move|test move red?(1.5708,0,0)@(1,0,0)\"\n" +
    "move=\"quartet move\" type=compound contents[4]=\"test move\" + \"test move red\" + \"test move\" + \"test move\"\n" +
    "move=\"test move\" type=simple "+
    "frames[2]=2.0 2.0 \n" +
    "figures[1]=\n" +
    "figure 0 color=ff00 {\n" +
    "frame 0 {\n" +
    "position=0.0 0.0 0.92 direction=0.0 0.0 0.0\n" +
    "spine=0.0 0.0 0.0 neck=0.0 0.0 0.0\n" +
    "left { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.523 0.0 0.392 elbow=1.047 wrist=0.392 0.0 }\n" +
    "right { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.392 0.0 elbow=0.523 wrist=0.0 0.392 }\n" +
    "}\n" +
    "frame 1 {\n" +
    "position=0.0 1.0 0.92 direction=1.5708 0.0 0.0\n" +
    "spine=0.0 0.0 0.0 neck=0.0 0.0 0.0\n" +
    "left { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.523 0.0 0.392 elbow=1.047 wrist=0.392 0.392 }\n" +
    "right { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.392 0.0 elbow=0.523 wrist=0.392 0.392 }\n" +
    "}\n" +
    "}\n" + 
    "move=\"test move red\" type=simple "+
    "frames[2]=2.0 2.0 \n" +
    "figures[1]=\n" +
    "figure 0 color=ff0000 {\n" +
    "frame 0 {\n" +
    "position=0.0 0.0 0.92 direction=0.0 0.0 0.0\n" +
    "spine=0.0 0.0 0.0 neck=0.0 0.0 0.0\n" +
    "left { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.523 0.0 0.392 elbow=1.047 wrist=0.392 0.0 }\n" +
    "right { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.392 0.0 elbow=0.523 wrist=0.0 0.392 }\n" +
    "}\n" +
    "frame 1 {\n" +
    "position=0.0 1.0 0.92 direction=1.5708 0.0 0.0\n" +
    "spine=0.0 0.0 0.0 neck=0.0 0.0 0.0\n" +
    "left { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.523 0.0 0.392 elbow=1.047 wrist=0.392 0.392 }\n" +
    "right { hip=0.0 0.0 0.0 knee=0.0 ankle=0.0 shoulder=0.0 0.0 humerus=0.0 0.392 0.0 elbow=0.523 wrist=0.392 0.392 }\n" +
    "}\n" +
    "}\n" + 
    "move=\"dual move\" type=simple " + savedPositions_;
}
